@@ -1,5 +1,6 @@
 from __future__ import annotations
 
+import itertools
 import typing as t
 from abc import ABC, abstractmethod
 from dataclasses import dataclass
@@ -15,16 +16,16 @@ from arg_services.nlp.v1 import nlp_pb2, nlp_pb2_grpc
 from mashumaro.mixins.dict import DataClassDictMixin
 from sentence_transformers import SentenceTransformer
 from spacy.language import Language as SpacyLanguage
-from spacy.tokens import Doc, DocBin
+from spacy.tokens import Doc, DocBin, Span, Token
 from thinc.types import Floats1d as SpacyVector
 from torch.cuda import is_available as is_cuda_available
-
-# import tensorflow_hub as hub
 from transformers import AutoModel, AutoTokenizer
 
 from nlp_service.similarity import SimilarityFactory as SimilarityFactory
 from nlp_service.types import ArrayLike, NumpyMatrix, NumpyVector
 
+# import tensorflow_hub as hub
+
 # https://spacy.io/usage/processing-pipelines#built-in
 spacy_components = (
     "tagger",
@@ -41,7 +42,7 @@ spacy_components = (
     "sentencizer",
     # tok2vec, transformer
 )
-custom_components = ("embedding_models", "similarity_method")
+custom_components = ("embeddings_factory", "similarity_factory")
 torch_device = "cuda" if is_cuda_available() else "cpu"
 
 # TODO: Extract spacy-specific code into its own file.
@@ -166,7 +167,7 @@ class SpacyModel(ModelBase):
         return doc.vector
 
 
-@SpacyLanguage.factory("embedding_models")
+@SpacyLanguage.factory("embeddings_factory")
 class EmbeddingsFactory:
     def __init__(self, nlp, name, models):
         self.models: list[ModelBase] = []
@@ -174,10 +175,10 @@ class EmbeddingsFactory:
         for model_dict in models:
             model = EmbeddingModel.from_dict(model_dict)
 
-            if model not in embedding_cache:
-                embedding_cache[model] = embedding_map[model.model_type](model)
+            if model not in model_cache:
+                model_cache[model] = embedding_map[model.model_type](model)
 
-            self.models.append(embedding_cache[model])
+            self.models.append(model_cache[model])
 
     def __call__(self, doc):
         if len(self.models) > 0:
@@ -193,11 +194,12 @@ class EmbeddingsFactory:
 
 
 SpacyKey = tuple[str, str, tuple[EmbeddingModel, ...]]
-spacy_cache: dict[SpacyKey, SpacyLanguage] = {}
-embedding_cache: dict[EmbeddingModel, ModelBase] = {}
+SpacyCache = t.Tuple[SpacyLanguage, dict[str, Doc]]
+spacy_cache: dict[SpacyKey, SpacyCache] = {}
+model_cache: dict[EmbeddingModel, ModelBase] = {}
 
 
-def _load_spacy(config: nlp_pb2.NlpConfig) -> SpacyLanguage:
+def _load_spacy(config: nlp_pb2.NlpConfig) -> SpacyCache:
     models = tuple(
         EmbeddingModel.from_protobuf(model) for model in config.embedding_models
     )
@@ -216,7 +218,7 @@ def _load_spacy(config: nlp_pb2.NlpConfig) -> SpacyLanguage:
 
         if models:
             nlp.add_pipe(
-                "embedding_models",
+                "embeddings_factory",
                 last=True,
                 config={"models": [model.to_dict() for model in models]},
             )
@@ -226,12 +228,12 @@ def _load_spacy(config: nlp_pb2.NlpConfig) -> SpacyLanguage:
             nlp_pb2.SIMILARITY_METHOD_UNSPECIFIED,
         ]:
             nlp.add_pipe(
-                "similarity_method",
+                "similarity_factory",
                 last=True,
                 config={"method": config.similarity_method},
             )
 
-        spacy_cache[key] = nlp
+        spacy_cache[key] = (nlp, {})
 
     return spacy_cache[key]
 
@@ -271,10 +273,12 @@ class NlpService(nlp_pb2_grpc.NlpServiceServicer):
                 ["model_type", "model_name"],
                 model,
                 ctx,
-                parent="embedding_models",
+                parent="embeddings_factory",
             )
 
-        nlp = _load_spacy(req.config)
+        # TODO: Cache not used due to the ability to enable/disable certain pipes
+        nlp, _ = _load_spacy(req.config)
+
         pipes_selection = {"disable": []}  # if empty, spacy will raise an exception
 
         if req.HasField("attributes") and not req.attributes.values:
@@ -285,7 +289,7 @@ class NlpService(nlp_pb2_grpc.NlpServiceServicer):
             pipes_selection = {"disable": list(req.disabled_pipes.values)}
 
         with nlp.select_pipes(**pipes_selection):
-            docs = t.cast(t.List[Doc], list(nlp.pipe(req.texts)))
+            docs = list(nlp.pipe(req.texts))
 
         if levels := req.embedding_levels:
             for doc in docs:
@@ -322,12 +326,14 @@ class NlpService(nlp_pb2_grpc.NlpServiceServicer):
             ctx,
         )
 
-        nlp = _load_spacy(req.config)
+        nlp, doc_cache = _load_spacy(req.config)
 
-        with nlp.select_pipes(enable=custom_components):
-            docs = nlp.pipe(req.texts)
+        if new_texts := [text for text in req.texts if text not in doc_cache]:
+            with nlp.select_pipes(enable=custom_components):
+                doc_cache.update(dict(zip(new_texts, nlp.pipe(new_texts))))
 
-        for doc in docs:
+        for text in req.texts:
+            doc = doc_cache[text]
             vector_res = nlp_pb2.VectorResponse()
 
             for level in req.embedding_levels:
@@ -371,16 +377,17 @@ class NlpService(nlp_pb2_grpc.NlpServiceServicer):
             ctx,
         )
 
-        nlp = _load_spacy(req.config)
-        text_tuples = ((x.text1, x.text2) for x in req.text_tuples)
-        texts1, texts2 = zip(*text_tuples)
+        nlp, doc_cache = _load_spacy(req.config)
+        texts = itertools.chain.from_iterable(
+            ((x.text1, x.text2) for x in req.text_tuples)
+        )
 
-        with nlp.select_pipes(enable=custom_components):
-            docs1 = nlp.pipe(texts1)
-            docs2 = nlp.pipe(texts2)
+        if new_texts := [text for text in texts if text not in doc_cache]:
+            with nlp.select_pipes(enable=custom_components):
+                doc_cache.update(dict(zip(new_texts, nlp.pipe(new_texts))))
 
         res.similarities.extend(
-            doc1.similarity(doc2) for doc1, doc2 in zip(docs1, docs2)
+            doc_cache[x.text1].similarity(doc_cache[x.text2]) for x in req.text_tuples
         )
 
         return res
